@@ -43,6 +43,7 @@
 #include <string.h>
 #include <linux/types.h>
 #include <stdint.h>
+#include <infiniband/verbs_api.h>
 
 #ifdef __cplusplus
 #include <limits>
@@ -133,6 +134,21 @@ enum ibv_atomic_cap {
 	IBV_ATOMIC_GLOB
 };
 
+struct ibv_alloc_dm_attr {
+	size_t length;
+	uint32_t log_align_req;
+	uint32_t comp_mask;
+};
+
+struct ibv_dm {
+	struct ibv_context *context;
+	int (*memcpy_to_dm)(struct ibv_dm *dm, uint64_t dm_offset,
+			    const void *host_addr, size_t length);
+	int (*memcpy_from_dm)(void *host_addr, struct ibv_dm *dm,
+			      uint64_t dm_offset, size_t length);
+	uint32_t comp_mask;
+};
+
 struct ibv_device_attr {
 	char			fw_ver[64];
 	__be64			node_guid;
@@ -230,6 +246,7 @@ enum ibv_rx_hash_fields {
 	IBV_RX_HASH_DST_PORT_TCP	= 1 << 5,
 	IBV_RX_HASH_SRC_PORT_UDP	= 1 << 6,
 	IBV_RX_HASH_DST_PORT_UDP	= 1 << 7,
+	IBV_RX_HASH_IPSEC_SPI		= 1 << 8,
 	IBV_RX_HASH_INNER		= (1UL << 31),
 };
 
@@ -290,6 +307,7 @@ struct ibv_device_attr_ex {
 	uint32_t		raw_packet_caps; /* Use ibv_raw_packet_caps */
 	struct ibv_tm_caps	tm_caps;
 	struct ibv_cq_moderation_caps  cq_mod_caps;
+	uint64_t max_dm_size;
 };
 
 enum ibv_mtu {
@@ -328,6 +346,7 @@ enum ibv_port_cap_flags {
 	IBV_PORT_SYS_IMAGE_GUID_SUP		= 1 << 11,
 	IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP	= 1 << 12,
 	IBV_PORT_EXTENDED_SPEEDS_SUP		= 1 << 14,
+	IBV_PORT_CAP_MASK2_SUP			= 1 << 15,
 	IBV_PORT_CM_SUP				= 1 << 16,
 	IBV_PORT_SNMP_TUNNEL_SUP		= 1 << 17,
 	IBV_PORT_REINIT_SUP			= 1 << 18,
@@ -341,6 +360,15 @@ enum ibv_port_cap_flags {
 	IBV_PORT_IP_BASED_GIDS			= 1 << 26
 };
 
+enum ibv_port_cap_flags2 {
+	IBV_PORT_SET_NODE_DESC_SUP		= 1 << 0,
+	IBV_PORT_INFO_EXT_SUP			= 1 << 1,
+	IBV_PORT_VIRT_SUP			= 1 << 2,
+	IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP	= 1 << 3,
+	IBV_PORT_LINK_WIDTH_2X_SUP		= 1 << 4,
+	IBV_PORT_LINK_SPEED_HDR_SUP		= 1 << 5,
+};
+
 struct ibv_port_attr {
 	enum ibv_port_state	state;
 	enum ibv_mtu		max_mtu;
@@ -362,7 +390,8 @@ struct ibv_port_attr {
 	uint8_t			active_speed;
 	uint8_t			phys_state;
 	uint8_t			link_layer;
-	uint8_t			reserved;
+	uint8_t			flags;
+	uint16_t		port_cap_flags2;
 };
 
 enum ibv_event_type {
@@ -632,7 +661,11 @@ enum ibv_rate {
 	IBV_RATE_25_GBPS  = 15,
 	IBV_RATE_100_GBPS = 16,
 	IBV_RATE_200_GBPS = 17,
-	IBV_RATE_300_GBPS = 18
+	IBV_RATE_300_GBPS = 18,
+	IBV_RATE_28_GBPS  = 19,
+	IBV_RATE_50_GBPS  = 20,
+	IBV_RATE_400_GBPS = 21,
+	IBV_RATE_600_GBPS = 22,
 };
 
 /**
@@ -777,7 +810,7 @@ struct ibv_wq_attr {
  * Receive Work Queue Indirection Table.
  * It's used in order to distribute incoming packets between different
  * Receive Work Queues. Associating Receive WQs with different CPU cores
- * allows to workload the traffic between different CPU cores.
+ * allows one to workload the traffic between different CPU cores.
  * The Indirection Table can contain only WQs of type IBV_WQT_RQ.
 */
 struct ibv_rwq_ind_table {
@@ -969,6 +1002,13 @@ struct ibv_qp_attr {
 	uint32_t		rate_limit;
 };
 
+struct ibv_qp_rate_limit_attr {
+	uint32_t	rate_limit;  /* in kbps */
+	uint32_t	max_burst_sz; /* total burst size in bytes */
+	uint16_t	typical_pkt_sz; /* typical send packet size in bytes */
+	uint32_t	comp_mask;
+};
+
 enum ibv_wr_opcode {
 	IBV_WR_RDMA_WRITE,
 	IBV_WR_RDMA_WRITE_WITH_IMM,
@@ -1348,6 +1388,7 @@ struct ibv_ah {
 enum ibv_flow_flags {
 	IBV_FLOW_ATTR_FLAGS_ALLOW_LOOP_BACK = 1 << 0,
 	IBV_FLOW_ATTR_FLAGS_DONT_TRAP = 1 << 1,
+	IBV_FLOW_ATTR_FLAGS_EGRESS = 1 << 2,
 };
 
 enum ibv_flow_attr_type {
@@ -1370,12 +1411,17 @@ enum ibv_flow_spec_type {
 	IBV_FLOW_SPEC_IPV4		= 0x30,
 	IBV_FLOW_SPEC_IPV6		= 0x31,
 	IBV_FLOW_SPEC_IPV4_EXT		= 0x32,
+	IBV_FLOW_SPEC_ESP		= 0x34,
 	IBV_FLOW_SPEC_TCP		= 0x40,
 	IBV_FLOW_SPEC_UDP		= 0x41,
 	IBV_FLOW_SPEC_VXLAN_TUNNEL	= 0x50,
+	IBV_FLOW_SPEC_GRE		= 0x51,
+	IBV_FLOW_SPEC_MPLS		= 0x60,
 	IBV_FLOW_SPEC_INNER		= 0x100,
 	IBV_FLOW_SPEC_ACTION_TAG	= 0x1000,
 	IBV_FLOW_SPEC_ACTION_DROP	= 0x1001,
+	IBV_FLOW_SPEC_ACTION_HANDLE	= 0x1002,
+	IBV_FLOW_SPEC_ACTION_COUNT	= 0x1003,
 };
 
 struct ibv_flow_eth_filter {
@@ -1439,6 +1485,18 @@ struct ibv_flow_spec_ipv6 {
 	struct ibv_flow_ipv6_filter mask;
 };
 
+struct ibv_flow_esp_filter {
+	uint32_t spi;
+	uint32_t seq;
+};
+
+struct ibv_flow_spec_esp {
+	enum ibv_flow_spec_type type;
+	uint16_t size;
+	struct ibv_flow_esp_filter val;
+	struct ibv_flow_esp_filter mask;
+};
+
 struct ibv_flow_tcp_udp_filter {
 	uint16_t dst_port;
 	uint16_t src_port;
@@ -1451,6 +1509,44 @@ struct ibv_flow_spec_tcp_udp {
 	struct ibv_flow_tcp_udp_filter mask;
 };
 
+struct ibv_flow_gre_filter {
+	/* c_ks_res0_ver field is bits 0-15 in offset 0 of a standard GRE header:
+	 * bit 0 - checksum present bit.
+	 * bit 1 - reserved. set to 0.
+	 * bit 2 - key present bit.
+	 * bit 3 - sequence number present bit.
+	 * bits 4:12 - reserved. set to 0.
+	 * bits 13:15 - GRE version.
+	 */
+	uint16_t c_ks_res0_ver;
+	uint16_t protocol;
+	uint32_t key;
+};
+
+struct ibv_flow_spec_gre {
+	enum ibv_flow_spec_type  type;
+	uint16_t  size;
+	struct ibv_flow_gre_filter val;
+	struct ibv_flow_gre_filter mask;
+};
+
+struct ibv_flow_mpls_filter {
+	/* The field includes the entire MPLS label:
+	 * bits 0:19 - label value field.
+	 * bits 20:22 - traffic class field.
+	 * bits 23 - bottom of stack bit.
+	 * bits 24:31 - ttl field.
+	 */
+	uint32_t label;
+};
+
+struct ibv_flow_spec_mpls {
+	enum ibv_flow_spec_type  type;
+	uint16_t  size;
+	struct ibv_flow_mpls_filter val;
+	struct ibv_flow_mpls_filter mask;
+};
+
 struct ibv_flow_tunnel_filter {
 	uint32_t tunnel_id;
 };
@@ -1473,6 +1569,18 @@ struct ibv_flow_spec_action_drop {
 	uint16_t  size;
 };
 
+struct ibv_flow_spec_action_handle {
+	enum ibv_flow_spec_type  type;
+	uint16_t  size;
+	const struct ibv_flow_action *action;
+};
+
+struct ibv_flow_spec_counter_action {
+	enum ibv_flow_spec_type  type;
+	uint16_t  size;
+	struct ibv_counters *counters;
+};
+
 struct ibv_flow_spec {
 	union {
 		struct {
@@ -1484,9 +1592,14 @@ struct ibv_flow_spec {
 		struct ibv_flow_spec_tcp_udp tcp_udp;
 		struct ibv_flow_spec_ipv4_ext ipv4_ext;
 		struct ibv_flow_spec_ipv6 ipv6;
+		struct ibv_flow_spec_esp esp;
 		struct ibv_flow_spec_tunnel tunnel;
+		struct ibv_flow_spec_gre gre;
+		struct ibv_flow_spec_mpls mpls;
 		struct ibv_flow_spec_action_tag flow_tag;
 		struct ibv_flow_spec_action_drop drop;
+		struct ibv_flow_spec_action_handle handle;
+		struct ibv_flow_spec_counter_action flow_count;
 	};
 };
 
@@ -1510,6 +1623,31 @@ struct ibv_flow {
 	uint32_t	   handle;
 };
 
+struct ibv_flow_action {
+	struct ibv_context *context;
+};
+
+enum ibv_flow_action_esp_mask {
+	IBV_FLOW_ACTION_ESP_MASK_ESN    = 1UL << 0,
+};
+
+struct ibv_flow_action_esp_attr {
+	struct ibv_flow_action_esp *esp_attr;
+
+	enum ibv_flow_action_esp_keymat	keymat_proto;
+	uint16_t		keymat_len;
+	void			*keymat_ptr;
+
+	enum ibv_flow_action_esp_replay replay_proto;
+	uint16_t                replay_len;
+	void                    *replay_ptr;
+
+	struct ibv_flow_action_esp_encap *esp_encap;
+
+	uint32_t		comp_mask; /* Use enum ibv_flow_action_esp_mask */
+	uint32_t		esn;
+};
+
 struct ibv_device;
 struct ibv_context;
 
@@ -1538,62 +1676,47 @@ struct ibv_device {
 	char			ibdev_path[IBV_SYSFS_PATH_MAX];
 };
 
+struct _compat_ibv_port_attr;
 struct ibv_context_ops {
-	int			(*query_device)(struct ibv_context *context,
-					      struct ibv_device_attr *device_attr);
-	int			(*query_port)(struct ibv_context *context, uint8_t port_num,
-					      struct ibv_port_attr *port_attr);
-	struct ibv_pd *		(*alloc_pd)(struct ibv_context *context);
-	int			(*dealloc_pd)(struct ibv_pd *pd);
-	struct ibv_mr *		(*reg_mr)(struct ibv_pd *pd, void *addr, size_t length,
-					  int access);
-	int			(*rereg_mr)(struct ibv_mr *mr,
-					    int flags,
-					    struct ibv_pd *pd, void *addr,
-					    size_t length,
-					    int access);
-	int			(*dereg_mr)(struct ibv_mr *mr);
+	void *(*_compat_query_device)(void);
+	int (*_compat_query_port)(struct ibv_context *context,
+				  uint8_t port_num,
+				  struct _compat_ibv_port_attr *port_attr);
+	void *(*_compat_alloc_pd)(void);
+	void *(*_compat_dealloc_pd)(void);
+	void *(*_compat_reg_mr)(void);
+	void *(*_compat_rereg_mr)(void);
+	void *(*_compat_dereg_mr)(void);
 	struct ibv_mw *		(*alloc_mw)(struct ibv_pd *pd, enum ibv_mw_type type);
 	int			(*bind_mw)(struct ibv_qp *qp, struct ibv_mw *mw,
 					   struct ibv_mw_bind *mw_bind);
 	int			(*dealloc_mw)(struct ibv_mw *mw);
-	struct ibv_cq *		(*create_cq)(struct ibv_context *context, int cqe,
-					     struct ibv_comp_channel *channel,
-					     int comp_vector);
+	void *(*_compat_create_cq)(void);
 	int			(*poll_cq)(struct ibv_cq *cq, int num_entries, struct ibv_wc *wc);
 	int			(*req_notify_cq)(struct ibv_cq *cq, int solicited_only);
-	void			(*cq_event)(struct ibv_cq *cq);
-	int			(*resize_cq)(struct ibv_cq *cq, int cqe);
-	int			(*destroy_cq)(struct ibv_cq *cq);
-	struct ibv_srq *	(*create_srq)(struct ibv_pd *pd,
-					      struct ibv_srq_init_attr *srq_init_attr);
-	int			(*modify_srq)(struct ibv_srq *srq,
-					      struct ibv_srq_attr *srq_attr,
-					      int srq_attr_mask);
-	int			(*query_srq)(struct ibv_srq *srq,
-					     struct ibv_srq_attr *srq_attr);
-	int			(*destroy_srq)(struct ibv_srq *srq);
+	void *(*_compat_cq_event)(void);
+	void *(*_compat_resize_cq)(void);
+	void *(*_compat_destroy_cq)(void);
+	void *(*_compat_create_srq)(void);
+	void *(*_compat_modify_srq)(void);
+	void *(*_compat_query_srq)(void);
+	void *(*_compat_destroy_srq)(void);
 	int			(*post_srq_recv)(struct ibv_srq *srq,
 						 struct ibv_recv_wr *recv_wr,
 						 struct ibv_recv_wr **bad_recv_wr);
-	struct ibv_qp *		(*create_qp)(struct ibv_pd *pd, struct ibv_qp_init_attr *attr);
-	int			(*query_qp)(struct ibv_qp *qp, struct ibv_qp_attr *attr,
-					    int attr_mask,
-					    struct ibv_qp_init_attr *init_attr);
-	int			(*modify_qp)(struct ibv_qp *qp, struct ibv_qp_attr *attr,
-					     int attr_mask);
-	int			(*destroy_qp)(struct ibv_qp *qp);
+	void *(*_compat_create_qp)(void);
+	void *(*_compat_query_qp)(void);
+	void *(*_compat_modify_qp)(void);
+	void *(*_compat_destroy_qp)(void);
 	int			(*post_send)(struct ibv_qp *qp, struct ibv_send_wr *wr,
 					     struct ibv_send_wr **bad_wr);
 	int			(*post_recv)(struct ibv_qp *qp, struct ibv_recv_wr *wr,
 					     struct ibv_recv_wr **bad_wr);
-	struct ibv_ah *		(*create_ah)(struct ibv_pd *pd, struct ibv_ah_attr *attr);
-	int			(*destroy_ah)(struct ibv_ah *ah);
-	int			(*attach_mcast)(struct ibv_qp *qp, const union ibv_gid *gid,
-						uint16_t lid);
-	int			(*detach_mcast)(struct ibv_qp *qp, const union ibv_gid *gid,
-						uint16_t lid);
-	void			(*async_event)(struct ibv_async_event *event);
+	void *(*_compat_create_ah)(void);
+	void *(*_compat_destroy_ah)(void);
+	void *(*_compat_attach_mcast)(void);
+	void *(*_compat_detach_mcast)(void);
+	void *(*_compat_async_event)(void);
 };
 
 struct ibv_context {
@@ -1608,12 +1731,11 @@ struct ibv_context {
 
 enum ibv_cq_init_attr_mask {
 	IBV_CQ_INIT_ATTR_MASK_FLAGS	= 1 << 0,
-	IBV_CQ_INIT_ATTR_MASK_RESERVED	= 1 << 1
 };
 
 enum ibv_create_cq_attr_flags {
 	IBV_CREATE_CQ_ATTR_SINGLE_THREADED = 1 << 0,
-	IBV_CREATE_CQ_ATTR_RESERVED = 1 << 1,
+	IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN  = 1 << 1,
 };
 
 struct ibv_cq_init_attr_ex {
@@ -1647,6 +1769,29 @@ struct ibv_parent_domain_init_attr {
 	uint32_t comp_mask;
 };
 
+struct ibv_counters_init_attr {
+	uint32_t	comp_mask;
+};
+
+struct ibv_counters {
+	struct ibv_context	*context;
+};
+
+enum ibv_counter_description {
+	IBV_COUNTER_PACKETS,
+	IBV_COUNTER_BYTES,
+};
+
+struct ibv_counter_attach_attr {
+	enum ibv_counter_description counter_desc;
+	uint32_t index; /* Desired location index of the counter at the counters object */
+	uint32_t comp_mask;
+};
+
+enum ibv_read_counters_flags {
+	IBV_READ_COUNTERS_ATTR_PREFER_CACHED = 1 << 0,
+};
+
 enum ibv_values_mask {
 	IBV_VALUES_MASK_RAW_CLOCK	= 1 << 0,
 	IBV_VALUES_MASK_RESERVED	= 1 << 1
@@ -1659,6 +1804,38 @@ struct ibv_values_ex {
 
 struct verbs_context {
 	/*  "grows up" - new fields go here */
+	int (*query_port)(struct ibv_context *context, uint8_t port_num,
+			  struct ibv_port_attr *port_attr,
+			  size_t port_attr_len);
+	int (*advise_mr)(struct ibv_pd *pd,
+			 enum ibv_advise_mr_advice advice,
+			 uint32_t flags,
+			 struct ibv_sge *sg_list,
+			 uint32_t num_sges);
+	struct ibv_mr *(*alloc_null_mr)(struct ibv_pd *pd);
+	int (*read_counters)(struct ibv_counters *counters,
+			     uint64_t *counters_value,
+			     uint32_t ncounters,
+			     uint32_t flags);
+	int (*attach_counters_point_flow)(struct ibv_counters *counters,
+					  struct ibv_counter_attach_attr *attr,
+					  struct ibv_flow *flow);
+	struct ibv_counters *(*create_counters)(struct ibv_context *context,
+						struct ibv_counters_init_attr *init_attr);
+	int (*destroy_counters)(struct ibv_counters *counters);
+	struct ibv_mr *(*reg_dm_mr)(struct ibv_pd *pd, struct ibv_dm *dm,
+				    uint64_t dm_offset, size_t length,
+				    unsigned int access);
+	struct ibv_dm *(*alloc_dm)(struct ibv_context *context,
+				   struct ibv_alloc_dm_attr *attr);
+	int (*free_dm)(struct ibv_dm *dm);
+	int (*modify_flow_action_esp)(struct ibv_flow_action *action,
+				      struct ibv_flow_action_esp_attr *attr);
+	int (*destroy_flow_action)(struct ibv_flow_action *action);
+	struct ibv_flow_action *(*create_flow_action_esp)(struct ibv_context *context,
+							  struct ibv_flow_action_esp_attr *attr);
+	int (*modify_qp_rate_limit)(struct ibv_qp *qp,
+				    struct ibv_qp_rate_limit_attr *attr);
 	struct ibv_pd *(*alloc_parent_domain)(struct ibv_context *context,
 					      struct ibv_parent_domain_init_attr *attr);
 	int (*dealloc_td)(struct ibv_td *td);
@@ -1729,6 +1906,56 @@ static inline struct verbs_context *verbs_get_ctx(struct ibv_context *ctx)
  */
 struct ibv_device **ibv_get_device_list(int *num_devices);
 
+/*
+ * When statically linking the user can set RDMA_STATIC_PROVIDERS to a comma
+ * separated list of provider names to include in the static link, and this
+ * machinery will cause those providers to be included statically.
+ *
+ * Linking will fail if this is set for dynamic linking.
+ */
+#ifdef RDMA_STATIC_PROVIDERS
+#define _RDMA_STATIC_PREFIX_(_1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11,     \
+			     _12, _13, _14, _15, ...)                          \
+	&verbs_provider_##_1, &verbs_provider_##_2, &verbs_provider_##_3,      \
+		&verbs_provider_##_4, &verbs_provider_##_5,                    \
+		&verbs_provider_##_6, &verbs_provider_##_7,                    \
+		&verbs_provider_##_8, &verbs_provider_##_9,                    \
+		&verbs_provider_##_10, &verbs_provider_##_11,                  \
+		&verbs_provider_##_12, &verbs_provider_##_13,                  \
+		&verbs_provider_##_14, &verbs_provider_##_15
+#define _RDMA_STATIC_PREFIX(arg)                                               \
+	_RDMA_STATIC_PREFIX_(arg, none, none, none, none, none, none, none,    \
+			     none, none, none, none, none, none, none)
+
+struct verbs_devices_ops;
+extern const struct verbs_device_ops verbs_provider_bnxt_re;
+extern const struct verbs_device_ops verbs_provider_cxgb3;
+extern const struct verbs_device_ops verbs_provider_cxgb4;
+extern const struct verbs_device_ops verbs_provider_hfi1verbs;
+extern const struct verbs_device_ops verbs_provider_hns;
+extern const struct verbs_device_ops verbs_provider_i40iw;
+extern const struct verbs_device_ops verbs_provider_ipathverbs;
+extern const struct verbs_device_ops verbs_provider_mlx4;
+extern const struct verbs_device_ops verbs_provider_mlx5;
+extern const struct verbs_device_ops verbs_provider_mthca;
+extern const struct verbs_device_ops verbs_provider_nes;
+extern const struct verbs_device_ops verbs_provider_ocrdma;
+extern const struct verbs_device_ops verbs_provider_qedr;
+extern const struct verbs_device_ops verbs_provider_rxe;
+extern const struct verbs_device_ops verbs_provider_vmw_pvrdma;
+extern const struct verbs_device_ops verbs_provider_all;
+extern const struct verbs_device_ops verbs_provider_none;
+void ibv_static_providers(void *unused, ...);
+
+static inline struct ibv_device **__ibv_get_device_list(int *num_devices)
+{
+	ibv_static_providers(NULL, _RDMA_STATIC_PREFIX(RDMA_STATIC_PROVIDERS),
+			     NULL);
+	return ibv_get_device_list(num_devices);
+}
+#define ibv_get_device_list(num_devices) __ibv_get_device_list(num_devices)
+#endif
+
 /**
  * ibv_free_device_list - Free list from ibv_get_device_list()
  *
@@ -1791,17 +2018,26 @@ int ibv_query_device(struct ibv_context *context,
  * ibv_query_port - Get port properties
  */
 int ibv_query_port(struct ibv_context *context, uint8_t port_num,
-		   struct ibv_port_attr *port_attr);
+		   struct _compat_ibv_port_attr *port_attr);
 
 static inline int ___ibv_query_port(struct ibv_context *context,
 				    uint8_t port_num,
 				    struct ibv_port_attr *port_attr)
 {
-	/* For compatibility when running with old libibverbs */
-	port_attr->link_layer = IBV_LINK_LAYER_UNSPECIFIED;
-	port_attr->reserved   = 0;
+	struct verbs_context *vctx = verbs_get_ctx_op(context, query_port);
 
-	return ibv_query_port(context, port_num, port_attr);
+	if (!vctx) {
+		int rc;
+
+		memset(port_attr, 0, sizeof(*port_attr));
+
+		rc = ibv_query_port(context, port_num,
+				    (struct _compat_ibv_port_attr *)port_attr);
+		return rc;
+	}
+
+	return vctx->query_port(context, port_num, port_attr,
+				sizeof(*port_attr));
 }
 
 #define ibv_query_port(context, port_num, port_attr) \
@@ -1819,6 +2055,12 @@ int ibv_query_gid(struct ibv_context *context, uint8_t port_num,
 int ibv_query_pkey(struct ibv_context *context, uint8_t port_num,
 		   int index, __be16 *pkey);
 
+/**
+ * ibv_get_pkey_index - Translate a P_Key into a P_Key index
+ */
+int ibv_get_pkey_index(struct ibv_context *context, uint8_t port_num,
+		       __be16 pkey);
+
 /**
  * ibv_alloc_pd - Allocate a protection domain
  */
@@ -1851,6 +2093,45 @@ static inline int ibv_destroy_flow(struct ibv_flow *flow_id)
 	return vctx->ibv_destroy_flow(flow_id);
 }
 
+static inline struct ibv_flow_action *
+ibv_create_flow_action_esp(struct ibv_context *ctx,
+			   struct ibv_flow_action_esp_attr *esp)
+{
+	struct verbs_context *vctx = verbs_get_ctx_op(ctx,
+						      create_flow_action_esp);
+
+	if (!vctx) {
+		errno = ENOSYS;
+		return NULL;
+	}
+
+	return vctx->create_flow_action_esp(ctx, esp);
+}
+
+static inline int
+ibv_modify_flow_action_esp(struct ibv_flow_action *action,
+			   struct ibv_flow_action_esp_attr *esp)
+{
+	struct verbs_context *vctx = verbs_get_ctx_op(action->context,
+						      modify_flow_action_esp);
+
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->modify_flow_action_esp(action, esp);
+}
+
+static inline int ibv_destroy_flow_action(struct ibv_flow_action *action)
+{
+	struct verbs_context *vctx = verbs_get_ctx_op(action->context,
+						      destroy_flow_action);
+
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->destroy_flow_action(action);
+}
+
 /**
  * ibv_open_xrcd - Open an extended connection domain
  */
@@ -1963,6 +2244,125 @@ struct ibv_comp_channel *ibv_create_comp_channel(struct ibv_context *context);
  */
 int ibv_destroy_comp_channel(struct ibv_comp_channel *channel);
 
+/**
+ * ibv_advise_mr - Gives advice about an address range in MRs
+ * @pd - protection domain of all MRs for which the advice is for
+ * @advice - type of advice
+ * @flags - advice modifiers
+ * @sg_list - an array of memory ranges
+ * @num_sge - number of elements in the array
+ */
+static inline int ibv_advise_mr(struct ibv_pd *pd,
+				enum ibv_advise_mr_advice advice,
+				uint32_t flags,
+				struct ibv_sge *sg_list,
+				uint32_t num_sge)
+{
+	struct verbs_context *vctx;
+
+	vctx = verbs_get_ctx_op(pd->context, advise_mr);
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->advise_mr(pd, advice, flags, sg_list, num_sge);
+}
+
+/**
+ * ibv_alloc_dm - Allocate device memory
+ * @context - Context DM will be attached to
+ * @attr - Attributes to allocate the DM with
+ */
+static inline
+struct ibv_dm *ibv_alloc_dm(struct ibv_context *context,
+			    struct ibv_alloc_dm_attr *attr)
+{
+	struct verbs_context *vctx = verbs_get_ctx_op(context, alloc_dm);
+
+	if (!vctx) {
+		errno = ENOSYS;
+		return NULL;
+	}
+
+	return vctx->alloc_dm(context, attr);
+}
+
+/**
+ * ibv_free_dm - Free device allocated memory
+ * @dm - The DM to free
+ */
+static inline
+int ibv_free_dm(struct ibv_dm *dm)
+{
+	struct verbs_context *vctx = verbs_get_ctx_op(dm->context, free_dm);
+
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->free_dm(dm);
+}
+
+/**
+ * ibv_memcpy_to/from_dm - copy to/from device allocated memory
+ * @dm - The DM to copy to/from
+ * @dm_offset - Offset in bytes from beginning of DM to start copy to/form
+ * @host_addr - Host memory address to copy to/from
+ * @length - Number of bytes to copy
+ */
+static inline
+int ibv_memcpy_to_dm(struct ibv_dm *dm, uint64_t dm_offset,
+		     const void *host_addr, size_t length)
+{
+	return dm->memcpy_to_dm(dm, dm_offset, host_addr, length);
+}
+
+static inline
+int ibv_memcpy_from_dm(void *host_addr, struct ibv_dm *dm,
+		       uint64_t dm_offset, size_t length)
+{
+	return dm->memcpy_from_dm(host_addr, dm, dm_offset, length);
+}
+
+/*
+ * ibv_alloc_null_mr - Allocate a null memory region.
+ * @pd - The protection domain associated with the MR.
+ */
+static inline
+struct ibv_mr *ibv_alloc_null_mr(struct ibv_pd *pd)
+{
+	struct verbs_context *vctx;
+
+	vctx = verbs_get_ctx_op(pd->context, alloc_null_mr);
+	if (!vctx) {
+		errno = ENOSYS;
+		return NULL;
+	}
+
+	return vctx->alloc_null_mr(pd);
+}
+
+/**
+ * ibv_reg_dm_mr - Register device memory as a memory region
+ * @pd - The PD to associated this MR with
+ * @dm - The DM to register
+ * @dm_offset - Offset in bytes from beginning of DM to start registration from
+ * @length - Number of bytes to register
+ * @access - memory region access flags
+ */
+static inline
+struct ibv_mr *ibv_reg_dm_mr(struct ibv_pd *pd, struct ibv_dm *dm,
+			     uint64_t dm_offset,
+			     size_t length, unsigned int access)
+{
+	struct verbs_context *vctx = verbs_get_ctx_op(pd->context, reg_dm_mr);
+
+	if (!vctx) {
+		errno = ENOSYS;
+		return NULL;
+	}
+
+	return vctx->reg_dm_mr(pd, dm, dm_offset, length, access);
+}
+
 /**
  * ibv_create_cq - Create a completion queue
  * @context - Context CQ will be attached to
@@ -2322,6 +2722,24 @@ ibv_open_qp(struct ibv_context *context, struct ibv_qp_open_attr *qp_open_attr)
 int ibv_modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr,
 		  int attr_mask);
 
+/**
+ * ibv_modify_qp_rate_limit - Modify a queue pair rate limit values
+ * @qp - QP object to modify
+ * @attr - Attributes to configure the rate limiting values of the QP
+ */
+static inline int
+ibv_modify_qp_rate_limit(struct ibv_qp *qp,
+			 struct ibv_qp_rate_limit_attr *attr)
+{
+	struct verbs_context *vctx;
+
+	vctx = verbs_get_ctx_op(qp->context, modify_qp_rate_limit);
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->modify_qp_rate_limit(qp, attr);
+}
+
 /**
  * ibv_query_qp - Returns the attribute list and current values for the
  *   specified QP.
@@ -2578,6 +2996,58 @@ static inline int ibv_is_qpt_supported(uint32_t caps, enum ibv_qp_type qpt)
 	return !!(caps & (1 << qpt));
 }
 
+static inline struct ibv_counters *ibv_create_counters(struct ibv_context *context,
+						       struct ibv_counters_init_attr *init_attr)
+{
+	struct verbs_context *vctx;
+
+	vctx = verbs_get_ctx_op(context, create_counters);
+	if (!vctx) {
+		errno = ENOSYS;
+		return NULL;
+	}
+
+	return vctx->create_counters(context, init_attr);
+}
+
+static inline int ibv_destroy_counters(struct ibv_counters *counters)
+{
+	struct verbs_context *vctx;
+
+	vctx = verbs_get_ctx_op(counters->context, destroy_counters);
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->destroy_counters(counters);
+}
+
+static inline int ibv_attach_counters_point_flow(struct ibv_counters *counters,
+						 struct ibv_counter_attach_attr *attr,
+						 struct ibv_flow *flow)
+{
+	struct verbs_context *vctx;
+
+	vctx = verbs_get_ctx_op(counters->context, attach_counters_point_flow);
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->attach_counters_point_flow(counters, attr, flow);
+}
+
+static inline int ibv_read_counters(struct ibv_counters *counters,
+				    uint64_t *counters_value,
+				    uint32_t ncounters,
+				    uint32_t flags)
+{
+	struct verbs_context *vctx;
+
+	vctx = verbs_get_ctx_op(counters->context, read_counters);
+	if (!vctx)
+		return ENOSYS;
+
+	return vctx->read_counters(counters, counters_value, ncounters, flags);
+}
+
 #ifdef __cplusplus
 }
 #endif
