@@ -34,11 +34,13 @@
 #define _HNS_ROCE_U_H
 
 #include <stddef.h>
+#include <endian.h>
 #include <util/compiler.h>
 
 #include <infiniband/driver.h>
 #include <util/udma_barrier.h>
 #include <infiniband/verbs.h>
+#include <ccan/bitmap.h>
 #include <ccan/container_of.h>
 
 #define HNS_ROCE_CQE_ENTRY_SIZE		0x20
@@ -84,7 +86,7 @@ enum {
 struct hns_roce_device {
 	struct verbs_device		ibv_dev;
 	int				page_size;
-	struct hns_roce_u_hw		*u_hw;
+	const struct hns_roce_u_hw	*u_hw;
 	int				hw_version;
 };
 
@@ -93,6 +95,21 @@ struct hns_roce_buf {
 	unsigned int			length;
 };
 
+/* the sw doorbell type; */
+enum hns_roce_db_type {
+	HNS_ROCE_QP_TYPE_DB,
+	HNS_ROCE_CQ_TYPE_DB,
+	HNS_ROCE_DB_TYPE_NUM
+};
+
+struct hns_roce_db_page {
+	struct hns_roce_db_page	*prev, *next;
+	struct hns_roce_buf	buf;
+	unsigned int		num_db;
+	unsigned int		use_cnt;
+	bitmap			*bitmap;
+};
+
 struct hns_roce_context {
 	struct verbs_context		ibv_ctx;
 	void				*uar;
@@ -110,6 +127,10 @@ struct hns_roce_context {
 	int				num_qps;
 	int				qp_table_shift;
 	int				qp_table_mask;
+
+	struct hns_roce_db_page		*db_list[HNS_ROCE_DB_TYPE_NUM];
+	pthread_mutex_t			db_list_mutex;
+
 	unsigned int			max_qp_wr;
 	unsigned int			max_sge;
 	int				max_cqe;
@@ -130,6 +151,7 @@ struct hns_roce_cq {
 	unsigned int			*set_ci_db;
 	unsigned int			*arm_db;
 	int				arm_sn;
+	unsigned long			flags;
 };
 
 struct hns_roce_srq {
@@ -188,25 +210,20 @@ struct hns_roce_qp {
 	unsigned int			sq_signal_bits;
 	struct hns_roce_wq		sq;
 	struct hns_roce_wq		rq;
+	uint32_t			*rdb;
+	uint32_t			*sdb;
 	struct hns_roce_sge_ex		sge;
 	unsigned int			next_sge;
 	int				port_num;
 	int				sl;
 
 	struct hns_roce_rinl_buf	rq_rinl_buf;
+	unsigned long			flags;
 };
 
 struct hns_roce_u_hw {
 	uint32_t hw_version;
-	int (*poll_cq)(struct ibv_cq *ibvcq, int ne, struct ibv_wc *wc);
-	int (*arm_cq)(struct ibv_cq *ibvcq, int solicited);
-	int (*post_send)(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
-			 struct ibv_send_wr **bad_wr);
-	int (*post_recv)(struct ibv_qp *ibvqp, struct ibv_recv_wr *wr,
-			 struct ibv_recv_wr **bad_wr);
-	int (*modify_qp)(struct ibv_qp *qp, struct ibv_qp_attr *attr,
-			 int attr_mask);
-	int (*destroy_qp)(struct ibv_qp *ibqp);
+	struct verbs_context_ops hw_ops;
 };
 
 static inline unsigned long align(unsigned long val, unsigned long align)
@@ -254,14 +271,20 @@ int hns_roce_u_free_pd(struct ibv_pd *pd);
 
 struct ibv_mr *hns_roce_u_reg_mr(struct ibv_pd *pd, void *addr, size_t length,
 				 int access);
-int hns_roce_u_rereg_mr(struct ibv_mr *mr, int flags, struct ibv_pd *pd,
+int hns_roce_u_rereg_mr(struct verbs_mr *mr, int flags, struct ibv_pd *pd,
 			void *addr, size_t length, int access);
-int hns_roce_u_dereg_mr(struct ibv_mr *mr);
+int hns_roce_u_dereg_mr(struct verbs_mr *mr);
+
+struct ibv_mw *hns_roce_u_alloc_mw(struct ibv_pd *pd, enum ibv_mw_type type);
+int hns_roce_u_dealloc_mw(struct ibv_mw *mw);
+int hns_roce_u_bind_mw(struct ibv_qp *qp, struct ibv_mw *mw,
+		       struct ibv_mw_bind *mw_bind);
 
 struct ibv_cq *hns_roce_u_create_cq(struct ibv_context *context, int cqe,
 				    struct ibv_comp_channel *channel,
 				    int comp_vector);
 
+int hns_roce_u_modify_cq(struct ibv_cq *cq, struct ibv_modify_cq_attr *attr);
 int hns_roce_u_destroy_cq(struct ibv_cq *cq);
 void hns_roce_u_cq_event(struct ibv_cq *cq);
 
@@ -277,7 +300,7 @@ void hns_roce_free_buf(struct hns_roce_buf *buf);
 
 void hns_roce_init_qp_indices(struct hns_roce_qp *qp);
 
-extern struct hns_roce_u_hw hns_roce_u_hw_v1;
-extern struct hns_roce_u_hw hns_roce_u_hw_v2;
+extern const struct hns_roce_u_hw hns_roce_u_hw_v1;
+extern const struct hns_roce_u_hw hns_roce_u_hw_v2;
 
 #endif /* _HNS_ROCE_U_H */
